from argparse import ArgumentParser, Namespace
import boto3
import json
import os
from sneks.git import get_repo_name

SNEKS_CONFIG = {}

class SneksParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument("-f", "--filename",
                          help="Filename of the config file.  Defaults to 'sneks.json'",
                          dest="filename",
                          default=None)
    def parse_args(self, *args, **kwargs):
        args = super().parse_args(*args, **kwargs)
        sneks_config = load_config(args.filename)
        return Namespace(sneks_config=sneks_config, **vars(args))

def load_config(filename=None):
    global SNEKS_CONFIG
    _filename = filename if filename else "sneks.json"
    if not os.path.exists(_filename) or not os.path.isfile(_filename):
        if filename:
            raise RuntimeError("Conf file '{}' doesn't exist!\n".format(filename))
    else:
        with open(_filename) as f:
            SNEKS_CONFIG.update(json.load(f))
    SNEKS_CONFIG["AWS_REGION"] = SNEKS_CONFIG.get("AWS_REGION") if SNEKS_CONFIG.get("AWS_REGION") else "us-east-1"
    SNEKS_CONFIG["AWS_PROFILE"] = SNEKS_CONFIG.get("AWS_PROFILE") if SNEKS_CONFIG.get("AWS_PROFILE") else "default"
    boto3.setup_default_session(region_name=SNEKS_CONFIG["AWS_REGION"], profile_name=SNEKS_CONFIG["AWS_PROFILE"])
    return SNEKS_CONFIG

def load_stack_info():
    global SNEKS_CONFIG
    repo_name = get_repo_name()
    cf = boto3.client("cloudformation")
    stacks = cf.describe_stacks()["Stacks"]
    build_stack_name = SNEKS_CONFIG.get("BUILD_STACK_NAME", repo_name + "-build")
    build_stack = [s for s in stacks if s["StackName"] == build_stack_name][0]
    build_stack_params = {x["ParameterKey"]:x["ParameterValue"] for x in build_stack["Parameters"]}
    deploy_stack_name = build_stack_params.get("ChildStackName")
    if not deploy_stack_name:
        deploy_stack_name = "{PackageName}-{PackageBranch}".format(**build_stack_params)
    deploy_stack = [s for s in stacks if s["StackName"] == deploy_stack_name][0]
    return {
        "build_stack_name":build_stack["StackName"],
        "deploy_stack_name":deploy_stack_name,
        "build_stack":build_stack,
        "deploy_stack":deploy_stack,
        "build_stack_outputs":{x['OutputKey']:x['OutputValue'] for x in build_stack.get("Outputs",[])},
        "deploy_stack_outputs":{x['OutputKey']:x['OutputValue'] for x in deploy_stack.get("Outputs",[])},
    }



#!/usr/bin/env python3

import base64
from botocore.exceptions import *
from boto3.dynamodb.conditions import Key, Attr, Or
from boto3.dynamodb.types import TypeSerializer
import boto3
import copy
from datetime import datetime
import decimal
import html
import inspect
import json
import logging
import os
import random
import traceback

VERSION_KEY = '__version__'

DATETIME_FORMAT = "datetime:%Y-%m-%dT%H:%M:%S.%fZ"

logger = logging.getLogger(__name__)
_TRACE_LEVEL = 1
def _TRACE(msg, *args, **kwargs):
    return logger.log(_TRACE_LEVEL, msg, *args, **kwargs)

def DECIMAL(*args, **kwargs):
    d = decimal.Decimal(*args, **kwargs)
    decimal.getcontext().clear_flags()
    return d

def _ensure_ddbsafe(d):
    if isinstance(d, dict):
        return {k:_ensure_ddbsafe(d[k]) for k in d}
    elif isinstance(d, list):
        return [_ensure_ddbsafe(e) for e in d]
    elif isinstance(d, decimal.Decimal):
        return float(d)
    elif isinstance(d, datetime):
        return d.strftime(DATETIME_FORMAT)
    elif isinstance(d, str) and "" == d:
        return None
    else:
        return d

def ensure_ddbsafe(d):
    return json.loads(json.dumps(_ensure_ddbsafe(d)), parse_float=decimal.Decimal)

def _fix_types(d):
    if isinstance(d, dict):
        return {k:_fix_types(d[k]) for k in d}
    elif isinstance(d, list):
        return [_fix_types(e) for e in d]
    elif isinstance(d, decimal.Decimal):
        return float(d)
    elif isinstance(d, str):
        if d.startswith("datetime:"):
            try:
                return datetime.strptime(d, DATETIME_FORMAT)
            except:
                pass
    return d

class BaseDynamoObject(dict):
    """
    Holder class for a bunch of class methods and stuff like that.
    """
    _SCHEMA_CACHE = None
    _TABLE_CACHE = None
    _CLASSNAME = None
    _REQUIRED_ATTRS = []
    _COMPOUND_ATTRS = {}
    _DEFAULT_ITEMS = []
    _META_ITEMS = []

    @classmethod
    def _from_dict(cls, d):
        return cls(d)

    @classmethod
    def _encode_nexttoken(cls, key):
        # TODO: replace the json step with something that'll work for any input
        # Convert to a string
        # This won't work for a lot of valid keys due to the json serialization step
        key = _fix_types(key)
        key = json.dumps(key)
        # base64 encode it to make it safer to handle
        key = base64.urlsafe_b64encode(key.encode("utf-8")).decode("utf-8")
        # strip the padding, as we can easily re-add it later
        key = key.replace("=","")
        return key

    @classmethod
    def _decode_nexttoken(cls, key):
        # TODO: replace the json step with something that'll work for any input
        # re-add the padding
        key = key + "=" * ((-1*len(key))%16)
        # convert if back from base64-encoded bytes to the underlying string
        key = base64.urlsafe_b64decode(key.encode("utf-8")).decode("utf-8")
        # load the string back into a dict
        # This won't work for a lot of valid keys due to the json serialization step
        key = json.loads(key)
        key = ensure_ddbsafe(key)
        return key

    @classmethod
    def _parse_items(cls, response):
        items = []
        for item in response.get("Items",[]):
            items.append(cls(dict(_fix_types(item))))
        return items

    @classmethod
    def _preprocess_search_params(cls, **kwargs):
        params = dict(kwargs)
        if params.get("ShufflePages", None):
            del params["ShufflePages"]
        if params.get("PageSize", None):
            if not params.get("Limit", None):
                params["Limit"] = params["PageSize"]
            del params["PageSize"]
        if params.get("MaxResults", None):
            if params["MaxResults"] > 0 and not params.get("Limit", None):
                # This is an arbitrary padding factor in case the user is using some sort of server-side filtering.
                # Limit sets the number of items evaluated, so the number returned may be smaller.
                # TODO: Either properly document this, or make it less janktastic.
                params["Limit"] = 5*params["MaxResults"]
            del params["MaxResults"]
        if params.get("NextToken", None) and not params.get("ExclusiveStartKey", None):
            params["ExclusiveStartKey"] = cls._decode_nexttoken(params["NextToken"])
        if "NextToken" in params:
            del params["NextToken"]
        if not params.get("KeyConditionExpression", None):
            hashname, rangename = cls._HASH_AND_RANGE_KEYS(index_name = params.get("IndexName", None))
            if params.get(hashname) and not params.get("HashKey", None):
                params["HashKey"] = params[hashname]
                del params[hashname]
            if params.get(rangename) and not params.get("RangeKey", None):
                params["RangeKey"] = params[rangename]
                del params[rangename]
            if params.get("HashKey", None) and not params.get("KeyConditionExpression", None):
                hkc = Key(hashname).eq(params["HashKey"])
                if params.get("RangeKey", None):
                    rk = params["RangeKey"]
                    if isinstance(rk,(list, tuple)):
                        rkc = getattr(Key(rangename),rk[0])(*rk[1:])
                    else:
                        rkc = Key(rangename).eq(rk_arg)
                    kce = hkc & rkc
                else:
                    kce = hkc
                params["KeyConditionExpression"] = kce
        if "HashKey" in params:
            del params["HashKey"]
        if "RangeKey" in params:
            del params["RangeKey"]
        return params

    @classmethod
    def _postprocess_search_results(cls, results):
        response = {
            "Items":cls._parse_items(results),
            "Count":results.get("Count",0),
            "ScannedCount":results.get("ScannedCount",0),
            "NextToken":None,
            "RawResponse":results
        }
        if results.get("LastEvaluatedKey", None):
            response["NextToken"] = cls._encode_nexttoken(results["LastEvaluatedKey"])
        return response

    @classmethod
    def _scanquery(cls, func_name, **kwargs):
        # Almost certainly a better way to have these share this code,
        # but it'd take literally **minutes** to figure it out and ain't nobody got time for that.
        if func_name not in ["scan","query"]:
            logger.error("Nice try, kiddo.")
            raise RuntimeError("Invalid search operation '{}' specified.".format(func_name))
        params = cls._preprocess_search_params(**kwargs)
        results = getattr(cls.TABLE(),func_name)(**params)
        return cls._postprocess_search_results(results)

    @staticmethod
    def _autopaginate_search(func, **kwargs):
        # Almost certainly a better way to have these share this code,
        # but it'd take literally **minutes** to figure it out and ain't nobody got time for that.
        max_results = kwargs.get("MaxResults", -1)
        item_count = 0
        finished = False
        shuffle_pages = kwargs.get("ShufflePages", False)
        response = func(**kwargs)
        while response and not finished:
            items = response.get("Items")
            if shuffle_pages:
                random.shuffle(items)
            for item in items:
                if finished:
                    break
                item_count += 1
                if max_results > 0 and item_count >= max_results:
                    finished = True
                yield item
            if response.get("NextToken") and not finished:
                response = func(NextToken=response.get("NextToken"), **kwargs)
            else:
                response = None

    @classmethod
    def scan(cls, **kwargs):
        return cls._scanquery("scan", **kwargs)

    @classmethod
    def query(cls, **kwargs):
        return cls._scanquery("query", **kwargs)

    @classmethod
    def scan_all(cls, **kwargs):
        return cls._autopaginate_search(cls.scan, **kwargs)

    @classmethod
    def query_all(cls, **kwargs):
        return cls._autopaginate_search(cls.query, **kwargs)

    @classmethod
    def count(cls, **kwargs):
        kwargs["Select"] = "COUNT"
        params = cls._preprocess_search_params(**kwargs)
        results = cls.TABLE().query(**params)
        return cls._postprocess_search_results(results)

    @classmethod
    def count_all(cls, **kwargs):
        response = cls.count(**kwargs)
        item_count = 0
        scanned_count = 0
        while response:
            item_count += response.get("Count",0)
            scanned_count += response.get("ScannedCount",0)
            if response.get("NextToken"):
                response = cls.count(NextToken=response.get("NextToken"), **kwargs)
            else:
                response = None
        return {
            "Count": item_count,
            "ScannedCount": scanned_count
        }

    @classmethod
    def load(cls, **kwargs):
        obj = cls.TABLE().get_item(Key=kwargs).get("Item", {})
        if obj:
            obj = _fix_types(obj)
            return cls(obj)
        return None

    @classmethod
    def SCHEMA(cls, use_cache=True):
        if cls._SCHEMA_CACHE and use_cache:
            return cls._SCHEMA_CACHE
        schema = cls._SCHEMA()
        cls._SCHEMA_CACHE = schema
        return schema

    @classmethod
    def _SCHEMA(cls, use_cache=True):
        raise NotImplementedError("Each subclass must implement this on their own.")

    @classmethod
    def TABLE_NAME(cls, use_cache=True):
        schema = cls.SCHEMA(use_cache=use_cache)
        return schema.get('TableName')

    @classmethod
    def CLASS_NAME(cls):
        return cls._CLASSNAME if cls._CLASSNAME else "{module}.{name}".format(module=cls.__module__, name=cls.__name__)

    @classmethod
    def TABLE(cls):
        if not cls._TABLE_CACHE:
            cls._TABLE_CACHE = boto3.resource('dynamodb').Table(cls.TABLE_NAME())
        return cls._TABLE_CACHE

    @classmethod
    def create_table(cls):
        boto3.client("dynamodb").create_table(**cls._SCHEMA())

    @classmethod
    def _get_required_attributes(cls):
        attrs = []
        hashkn, rangekn = cls._HASH_AND_RANGE_KEYS()
        if hashkn:
            attrs.append(hashkn)
        if rangekn:
            attrs.append(rangekn)
        attrs.extend(cls._REQUIRED_ATTRS)
        return attrs

    @classmethod
    def _HASH_AND_RANGE_KEYS(cls, index_name=None):
        schema = cls._SCHEMA()
        key_schema = schema['KeySchema']
        if index_name:
            gsis = schema.get("GlobalSecondaryIndexes", [])
            matches = [gsi for gsi in gsis if gsi["IndexName"] == index_name]
            if len(matches) == 0:
                raise RuntimeError("No index with the name '{index_name}' found!".format(index_name=index_name))
            key_schema = matches[0]["KeySchema"]
        hash = [h['AttributeName'] for h in key_schema if h['KeyType']=='HASH'][0]
        ranges = [r['AttributeName'] for r in key_schema if r['KeyType']=='RANGE']
        range = ranges[0] if ranges else None
        return hash, range

    @classmethod
    def _get_class_relation_map(cls, obj):
        return {'class':cls.CLASS_NAME(), 'key':obj._get_key_dict()}

    @classmethod
    def _add_compound_attr(cls, attrname, attrfunc, save=False):
        cls._COMPOUND_ATTRS[attrname] = {"func":attrfunc,"save":save}

    @classmethod
    def _remove_compound_attr(cls, attrname):
        if attrname in cls._COMPOUND_ATTRS:
            del cls._COMPOUND_ATTRS[attrname]

class DynamoObject(BaseDynamoObject):
    '''
    Base class for all DynamoDB-storable objects.  Cannot itself be instantiated.

    The only thing that a subclass is required to implement is the classmethod SCHEMA, and it must return a dict that can be passed to client.create_table(**schema) and succeed.

    Constructor args:

    :param kwargs: Keys for an object, and any attributes to attach to that object.
    :rtype: DynamoObject
    '''
    def __init__(self, mapping={}, **kwargs):
        super().__init__(mapping, **kwargs)

    def __getitem__(self, key):
        if key in self._META_ITEMS:
            return self._META_ITEMS[key](self)
        try:
            return dict.__getitem__(self, key)
        except:
            if key in self._DEFAULT_ITEMS:
                _default = self._DEFAULT_ITEMS[key]
                return _default() if callable(_default) else _default
            raise

    def __setitem__(self, key, val):
        if key in self._META_ITEMS:
            # May want to make this an error at some point, but for now just make it a no-op for testing.
            # raise RuntimeError("Setting an explicit value for a meta-item ('{}') is not allowed.".format(key))
            return
        dict.__setitem__(self, key, val)

    def _my_hash_and_range(self):
        hash_keyname, range_keyname = self.__class__._HASH_AND_RANGE_KEYS()
        hash_key = self[hash_keyname]
        range_key = self[range_keyname] if range_keyname else None
        return hash_key, range_key

    def _get_key_dict(self, dictionary=None):
        hash_keyname, range_keyname = self.__class__._HASH_AND_RANGE_KEYS()
        keys = {}
        dictionary = dictionary if dictionary else self
        for k in (hash_keyname, range_keyname):
            if k and k in dictionary.keys():
                # I'm explicitly bypassing the getter here in the off chance either hash or range is a foreign key
                keys[k] = dictionary[k]
        return keys

    def save(self, force=False, save_if_missing=True, save_if_existing=True):
        _TRACE("DynamoObject.save reached")
        if not save_if_missing and not save_if_existing:
            raise RuntimeError("At least one of save_if_missing and save_if_existing must be true.")

        old_version = DECIMAL(self.get(VERSION_KEY, -1))
        create_condition = Attr(VERSION_KEY).not_exists()
        if force:
            update_condition = Attr(VERSION_KEY).exists()
        else:
            update_condition = Attr(VERSION_KEY).eq(old_version)
        CE = None
        if force and save_if_missing and save_if_existing:
            pass
        elif save_if_missing and save_if_existing:
            CE = Or(create_condition, update_condition)
        elif save_if_existing:
            CE = update_condition
        else:
            # If we're here, we know that create_condition=True
            CE = create_condition
        try:
            self[VERSION_KEY] = old_version + 1
            if CE:
                self._store(CE)
            else:
                self._store()
            return self
        except ClientError as e:
            self[VERSION_KEY] = old_version
            raise e

    def modify(self, force=False):
        return self.save(force=force, save_if_existing=True, save_if_missing=False)

    def create(self):
        return self.save(force=force, save_if_existing=False, save_if_missing=True)

    def _store(self, CE=None):
        _TRACE("DynamoObject._store reached")
        dict_to_save = copy.deepcopy(self)
        required = self._get_required_attributes()
        missing = [r for r in required if not dict_to_save.get(r, None)]
        if missing:
            raise RuntimeError('The following attributes are missing and must be added before saving: '+', '.join(missing))
        dict_to_save = ensure_ddbsafe(dict_to_save)
        if CE:
            self.__class__.TABLE().put_item(Item=dict_to_save, ConditionExpression=CE)
        else:
            self.__class__.TABLE().put_item(Item=dict_to_save)
        _TRACE("DynamoObject._store returning")
        return self

    def delete(self, CE=None):
        if CE:
            return self.__class__.TABLE().delete_item(Key=self._get_key_dict(), ConditionExpression=CE)
        else:
            return self.__class__.TABLE().delete_item(Key=self._get_key_dict())

    # def load(self):
    #     new_me = self.__class__.TABLE().get_item(Key=self._get_key_dict()).get("Item", {})
    #     return self.__class__.__init__(new_me)

    def reload(self):
        '''
        Reloads the item's attributes from DynamoDB, replacing whatever's currently in the object.
        '''
        new_me = self.load()
        for k in list(self.keys()):
            if k not in new_me:
                del self[k]
        self.update(new_me)
        return self

class CFObject(DynamoObject):
    '''
    Base class for objects that are based on tables created in a CloudFormation stack.
    '''

    _CF_STACK_NAME = None
    _CF_LOGICAL_NAME = None
    _CF_CLIENT = boto3.client('cloudformation')
    _CF_TEMPLATE = None
    _CF_RESOURCES = {}

    @classmethod
    def _set_cf_info(cls, cf_stack_name=None, cf_logical_name=None):
        if cf_stack_name:
            cls._CF_STACK_NAME = cf_stack_name
        if cf_logical_name:
            cls._LOGICAL_NAME = cf_logical_name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_stack_name(cls, stack_name=None):
        stack_name = stack_name if stack_name else cls._CF_STACK_NAME
        if not stack_name:
            raise RuntimeError("Stack name not set!")
        return stack_name

    @classmethod
    def _get_stack_and_logical_names(cls, stack_name=None, logical_name=None):
        stack_name = stack_name if stack_name else cls._CF_STACK_NAME
        logical_name = logical_name if logical_name else cls._CF_LOGICAL_NAME
        if not stack_name or not logical_name:
            raise RuntimeError("Stack name or logical name (or both) not set!")
        return stack_name, logical_name

    @classmethod
    def _describe_stack_resource(cls, stack_name=None, logical_name=None):
        stack_name, logical_name = cls._get_stack_and_logical_names(stack_name=stack_name, logical_name=logical_name)
        if logical_name not in cls._CF_RESOURCES:
            logging.warn("Cache miss loading _CF_RESOURCE for class {}".format(cls))
            response = cls._CF_CLIENT.describe_stack_resource(StackName=stack_name, LogicalResourceId=logical_name)
            if not response or "StackResourceDetail" not in response:
                raise RuntimeError("Resource does not exist!")
            cls._CF_RESOURCES[logical_name] = response["StackResourceDetail"]
        else:
            logging.info("Cache hit loading _CF_RESOURCE for class {}".format(cls))
        return cls._CF_RESOURCES[logical_name]

    @classmethod
    def _get_physical_resource_id(cls, stack_name=None, logical_name=None):
        return cls._describe_stack_resource(stack_name=stack_name, logical_name=logical_name)["PhysicalResourceId"]

    @classmethod
    def _get_template(cls, stack_name=None):
        stack_name = cls._get_stack_name(stack_name=stack_name)
        if not getattr(cls, "_CF_TEMPLATE"):
            logging.warn("Cache miss loading _CF_TEMPLATE for class {}".format(cls))
            try:
                template = cls._CF_CLIENT.get_template(StackName=stack_name)["TemplateBody"]
                setattr(cls, "_CF_TEMPLATE", template)
            except ValidationError:
                raise RuntimeError("Unable to retrieve template for stack {}, likely due to it not existing.".format(stack_name))
        else:
            logging.info("Cache hit loading _CF_TEMPLATE for class {}".format(cls))
        return cls._CF_TEMPLATE

    @classmethod
    def _clear_cf_cache(cls):
        setattr(cls, "_CF_TEMPLATE", None)
        setattr(cls, "_CF_RESOURCES", {})

    @classmethod
    def _SCHEMA(cls):
        stack_name, logical_name = cls._get_stack_and_logical_names()
        template = cls._get_template(stack_name=stack_name)
        resources = template["Resources"]
        if logical_name not in resources:
            raise RuntimeError("Stack doesn't contain a table with the given logical name!")
        resource = resources[logical_name]
        table_type = "AWS::DynamoDB::Table"
        if not table_type == resource["Type"]:
            raise RuntimeError("Logical resource {} in stack {} is of type '{}', not type '{}'".format(logical_name, stack_name, resource["Type"], table_type))
        properties = dict(resource["Properties"])
        properties["TableName"] = cls._get_physical_resource_id()
        return properties

    @classmethod
    def _get_class_relation_map(cls, obj):
        stack_name, logical_name = obj._get_stack_and_logical_names()
        return {'class':cls.CLASS_NAME(), '_cf_stack_name':stack_name, '_cf_logical_name':logical_name}

    @classmethod
    def lazysubclass(cls, stack_name=None, logical_name=None):
        '''
        Returns a class that inherits from this one, with the given default stack and logical names.
        If you just want to have a new object type with no fancy features added, this makes it a one-liner.

        :rtype: Class that inherits from cls
        '''
        class LazyObject(cls):
            _CF_STACK_NAME = stack_name if stack_name else cls._CF_STACK_NAME
            _CF_LOGICAL_NAME = logical_name if logical_name else cls._CF_LOGICAL_NAME
            _CLASSNAME = cls.CLASS_NAME()
        return LazyObject

class DataField(object):
    def __init__(self, key, **kwargs):
        self.key = key
        self.pretty = kwargs.get("pretty",key.replace("_"," ").title())
        self.hidden = kwargs.get("hidden",False)
        self.immutable = kwargs.get("immutable",False)
        self.data_type = kwargs.get("data_type","text")
        self.data_width = kwargs.get("data_width",50)
        self.data_height = kwargs.get("data_height",1)
        self.default = kwargs.get("default","")
        self.to_str = kwargs.get("to_str",lambda x:x)
        self.value_generator = kwargs.get("value_generator",None)
        self.validator = kwargs.get("validator",lambda x: True)
        pass

    def extract_value(self, d):
        return d.get(self.key, self.default)

    def form_format(self, d, immutable=False):
        value = self.extract_value(d)
        if self.hidden:
            value = html.escape(self.to_str(value) if value else "")
            response = '<input type="hidden" name="{key}" id="{key}_field" value="{value}"{immutable}>'
        else:
            response = '<label for="{key}">{pretty}: </label>\n'
            if self.data_type == "textarea":
                value = html.escape(self.to_str(value) if value else "")
                response += '<br><textarea name="{key}" id="{key}_field" rows="{data_height}" cols="{data_width}"{immutable}>{value}</textarea>'
            elif self.data_type == "datetime":
                if isinstance(value, datetime):
                    value = value.strftime("%Y-%m-%dT%H:%M")
                response += '<input type="datetime-local" name="{key}" id="{key}_field" value="{value}"{immutable}>'
            elif self.data_type == "int":
                value = int(value) if value else 0
                response += '<input type="number" name="{key}" id="{key}_field" value="{value}"{immutable}>'
            elif self.data_type == "boolean":
                value = "checked" if value else ""
                response += '<input type="checkbox" name="{key}" id="{key}_field" {value}'
                if immutable or self.immutable:
                    response += ' onclick="return false;"'
                response += '>'
            else:
                value = html.escape(self.to_str(value) if value else "")
                response += '<input type="text" name="{key}" id="{key}_field" value="{value}"{immutable}>'
            response += "<br>\n"
        return response.format(
            key=self.key,
            pretty=self.pretty,
            data_height=self.data_height,
            data_width=self.data_width,
            immutable=" readonly" if (immutable or self.immutable) else "",
            value=value
        )

class StructuredObject(DynamoObject):
    _HIDDEN_KEYS = [VERSION_KEY]
    _DATA_FIELDS = []

    @classmethod
    def skema(cls):
        return {x.key:x for x in cls._DATA_FIELDS}

    @classmethod
    def skema_keys(cls):
        return [k.key for k in cls._DATA_FIELDS]

    def all_keys(self):
        sk = self.skema_keys()
        return sk + [x for x in self if x not in sk]

    @classmethod
    def skeme(cls, k):
        return cls.skema().get(k, DataField(k))

    @classmethod
    def from_body(cls, body):
        raw_obj = json.loads(body)
        obj = {k:skeme(k).preprocessor(raw_obj[k]) for k in raw_obj}
        return cls(obj)

    def save(self, *args, **kwargs):
        for k in self.all_keys():
            skeme = self.skeme(k)
            if skeme.value_generator and not self.get(k):
                self[k] = skeme.value_generator()
            v = self.get(k)
            if not skeme.validator(v):
                raise RuntimeError('Invalid value provided for key "{}"!'.format(k))
        return super().save(*args, **kwargs)

    def data_form(self, immutable=False):
        keys = self.skema_keys()
        blocks = []
        for k in keys:
            v = self.skeme(k).form_format(self, immutable=immutable)
            if v:
                blocks.append(v)
        return '''<form>\n{}\n</form>'''.format("\n".join(blocks))



from decimal import Decimal
from math import radians, degrees, cos, sin, asin, sqrt, fabs, log, tan, pi, atan2

def bearing(lat1, lon1, lat2, lon2):
    lat1, lon1, lat2, lon2 = map(float, [lat1, lon1, lat2, lon2])
    startLat, startLong, endLat, endLong = map(radians, [lat1, lon1, lat2, lon2])
    dPhi = log(tan(endLat/2.0+pi/4.0)/tan(startLat/2.0+pi/4.0))
    if abs(dLong) > pi:
        if dLong > 0.0:
            dLong = -(2.0 * pi - dLong)
        else:
            dLong = (2.0 * pi + dLong)
    bearing = (degrees(atan2(dLong, dPhi)) + 360.0) % 360.0
    return bearing

def haversine(lat1, lon1, lat2, lon2):
    lat1, lon1, lat2, lon2 = map(float, [lat1, lon1, lat2, lon2])
    # https://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))
    r = 6371000 # Radius of earth in meters. Use 3959 for miles or 20903520 for feet
    return c * r

def minimize_gps(point, distance=1):
    latitude = point["lat"]
    longitude = point["lng"]
    lat_str = str(latitude)
    lng_str = str(longitude)
    _lat_str = lat_str
    _lng_str = lng_str
    while haversine(latitude, longitude, float(_lat_str), float(_lng_str)) < distance:
        lat_str = _lat_str
        lng_str = _lng_str
        if len(_lat_str.split(".")[1]) >= len(_lng_str.split(".")[1]):
            _lat_str = _lat_str[:-1]
        if len(_lat_str.split(".")[1]) <= len(_lng_str.split(".")[1]):
            _lng_str = _lng_str[:-1]
    return {"lat":Decimal(lat_str), "lng":Decimal(lng_str)}



import traceback
from sneks import snekjson as json
from functools import update_wrapper
from sneks.sam import ui_stuff, response_core

returns_html = ui_stuff.loader_for

def returns_json(func):
    def newfunc(event, *args, **kwargs):
        response = func(event, *args, **kwargs)
        if ui_stuff.is_response(response):
            return response
        return response_core.make_response(body=json.dumps(response, separators=(',',':')), headers = {"Content-Type": "application/json"})
    update_wrapper(newfunc, func)
    return newfunc

def register_path(chain_name, path_re, *args, **kwargs):
    def newfunc(func):
        response_core.register_path_matcher(chain_name, path_re, func, *args, **kwargs)
        return func
    return newfunc



import boto3
from boto3.dynamodb.conditions import Key, Attr
import sneks.snekjson as json
from decimal import Decimal
from datetime import datetime, timedelta
import traceback

STRFTIME_STRING = "%Y%m%d %H:%M:%S.%f"

def make_json_safe(item):
    if isinstance(item, list):
        item = [make_json_safe(e) for e in item]
    if isinstance(item, dict):
        item = {k:make_json_safe(item[k]) for k in item}
    if isinstance(item, Decimal):
        item = float(item)
    if isinstance(item, datetime):
        item = item.strftime(STRFTIME_STRING)
    return item

def dumps(obj, *args, **kwargs):
    return json.dumps(make_json_safe(obj), *args, **kwargs)

def make_ddb_safe(item):
    if isinstance(item, list):
        item = [make_ddb_safe(e) for e in item]
    if isinstance(item, dict):
        item = {k:make_ddb_safe(item[k]) for k in item}
    if isinstance(item, float):
        item = Decimal(item)
    if isinstance(item, datetime):
        item = item.strftime(STRFTIME_STRING)
    if item is None:
        item = "null"
    if item == "":
        item = json.dumps("")
    return item

def deepload(s):
    while True:
        try:
            s = json.loads(s)
        except:
            break
    if isinstance(s, list):
        s = [deepload(x) for x in s]
    if isinstance(s, dict):
        s = {k:deepload(s[k]) for k in s}
    return s

def ddb_save(table, item, **kwargs):
    item = make_ddb_safe(item)
    try:
        table.put_item(Item=item, **kwargs)
    except:
        print("Error saving the following object:")
        print(dumps(item))
        traceback.print_exc()
        raise

def ddb_load(table, key, **kwargs):
    item = None
    try:
        item = table.get_item(Key=key, **kwargs)["Item"]
    except:
        traceback.print_exc()
    return item

def ddb_query(table, key, value, **kwargs):
    # Todo: turn into a generator that auto-paginates.
    return table.query(Select='ALL_ATTRIBUTES', KeyConditionExpression=Key(key).eq(value), **kwargs)["Items"]

def table_from_env(varname):
    return boto3.resource("dynamodb").Table(os.environ[varname])



#!/usr/bin/env python3

import os
from setuptools import setup

MAJOR_VERSION='0'
MINOR_VERSION='0'
PATCH_VERSION='20'

VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION)

packages = ['sneks','sneks.sam','sneks.ddb']
package_dir = {p: 'src/' + p.replace('.','/') for p in packages}

scriptdir = "scripts"
script_files = [os.path.join(scriptdir, f) for f in os.listdir(scriptdir) if f[-1] not in "~#" and f[0] not in "~#."]
scripts = [f for f in script_files if os.access(f, os.X_OK)]

def main():
    setup(
        name = 'sneks',
        packages = packages,
        package_dir = package_dir,
        version = VERSION,
        description = 'Basic python utilities.',
        author = 'Steve Norum',
        author_email = 'sn@drunkenrobotlabs.org',
        url = 'https://github.com/stevenorum/sneks',
        download_url = 'https://github.com/stevenorum/sneks/archive/{}.tar.gz'.format(VERSION),
        keywords = ['python'],
        classifiers = [],
        install_requires=[
            'beautifulsoup4',
            'boto3',
            'jinja2'
        ],
        scripts = scripts,
        test_suite = 'tests'
    )

if __name__ == "__main__":
    main()



from bs4 import BeautifulSoup
import copy
import json
import re
from urllib.parse import urlencode, urlparse, parse_qs

def make_response(body, code=200, headers={}, base64=False, prettify_html=True):
    _headers = {"Content-Type": "text/html"}
    if isinstance(body, (list,dict)):
        body = json.dumps(body, separators=(',',':'))
        _headers = {"Content-Type": "application/json"}
    _headers.update(headers)
    if prettify_html and _headers.get("Content-Type") == "text/html":
        body = BeautifulSoup(body).prettify()
    return {
        "body": body,
        "statusCode": code,
        "headers": _headers,
        "isBase64Encoded": base64
    }

def redirect(target_url, temporary=True, headers={}, qs={}):
    print("Redirecting to {} with additional qs {}".format(target_url, qs))
    if qs:
        parsed = urlparse(target_url)
        eqs = parse_qs(parsed.query)
        newqs = set()
        for k in eqs:
            for e in eqs[k]:
                newqs.add(urlencode({k:e}))
        for k in qs:
            newqs.add(urlencode({k:qs[k]}))
        full_qs = "&".join(list(newqs))
        target_url = parsed.netloc + parsed.path
        if parsed.scheme:
            target_url = parsed.scheme + "://" + target_url
        if parsed.params:
            target_url += ";" + parsed.params
        if full_qs:
            target_url += "?" + full_qs
        if parsed.fragment:
            target_url += "#" + parsed.fragment
    _headers = {"Location": target_url}
    _headers.update(headers)
    print("Redirect headers: {}".format(_headers))
    return make_response(
        body = "",
        code = 303 if temporary else 301,
        headers = _headers
    )

class EventMatcher(object):
    def __init__(self, response_function, default_kwargs={}, matcher_function=None, preprocessor_functions=[]):
        self.response_function = response_function
        self.kwargs = default_kwargs
        preprocessor_functions = preprocessor_functions if isinstance(preprocessor_functions, (list, tuple)) else [preprocessor_functions]
        self.preprocessor_functions = preprocessor_functions if preprocessor_functions else getattr(self, "preprocessor_functions", [])
        self.matcher_function = matcher_function if matcher_function else getattr(self, "matcher_function", lambda *args, **kwargs: (False,None))

    def match_event(self, event):
        resp = self.matcher_function(event)
        if len(resp) != 2 or not resp[0]:
            return None, None
        kwargs = copy.deepcopy(self.kwargs)
        kwargs.update(resp[1])
        kwargs["event"] = event if "event" not in kwargs else kwargs.get("event")
        for processor_func in self.preprocessor_functions:
            kwargs = processor_func(kwargs)
        return (self.response_function, kwargs)

    def handle_event(self, event):
        match = self.match_event(event)
        if match and match[0]:
            function = match[0]
            kwargs = match[1]
            return function(**kwargs)
        return None

class PathMatcher(EventMatcher):
    def __init__(self, regex, function, default_kwargs={}, preprocessor_functions=[]):
        super().__init__(function, default_kwargs, preprocessor_functions=preprocessor_functions)
        self.matcher = re.compile(regex)

    def matcher_function(self, event):
        path = event.get("path")
        response = self.matcher.match(path)
        if response:
            return (True, response.groupdict())
        return False, None

class ListMatcher(EventMatcher):
    def __init__(self, matchers):
        self.matchers = matchers

    def match_event(self, event):
        for matcher in self.matchers:
            resp = matcher.match_event(event)
            if resp[0]:
                return resp
        return None, None

class ResponseException(Exception):
    def __init__(self, response):
        self.response = response

class ApiException(ResponseException):
    def __init__(self, data={}, code=500):
        self.response = make_response(body=data, code=code)

MATCHERS = {}

def register_path_matcher(chain_name, *args, **kwargs):
    global MATCHERS
    MATCHERS[chain_name] = MATCHERS.get(chain_name, [])
    MATCHERS[chain_name].append((args, kwargs))

def get_matchers(chain_name, **kwargs):
    global MATCHERS
    matchers = []
    for link in MATCHERS.get(chain_name, []):
        _args = link[0]
        _kwargs = copy.deepcopy(kwargs)
        _kwargs.update(link[1])
        matchers.append(PathMatcher(*_args, **_kwargs))
    return matchers



#!/usr/bin/env python3

import base64
from botocore.exceptions import *
from boto3.dynamodb.conditions import Key, Attr, Or
from boto3.dynamodb.types import TypeSerializer
import boto3
import copy
from datetime import datetime
import decimal
import inspect
import json
import logging
import os
import traceback

VERSION_KEY = 'version_toco_'

JSON_CLASS = '_class_toco'
JSON_FKEY = '_fkey_toco'

RELATION_SUFFIX = '_rel_toco_'
FKEY_PREFIX = 'toco_fkey='

FKEY_EMPTY_STRING = FKEY_PREFIX + "EMPTY-STRING"
CONSTANT_FKEYS = {FKEY_EMPTY_STRING: ""}

DATETIME_FORMAT = "datetime:%Y-%m-%dT%H:%M:%S.%fZ"

logger = logging.getLogger(__name__)

def load_python_class_if_applicable(value):
    if value and isinstance(value, str) and value.startswith("datetime:"):
        try:
            return datetime.strptime(value, DATETIME_FORMAT)
        except:
            pass
    return value

def get_class(clazzname):
    '''
    Dynamically retrieve a class from its name.

    (From http://stackoverflow.com/questions/547829/how-to-dynamically-load-a-python-class)

    :param clazzname: Name of the class to load.
    :rtype: Class
    '''
    components = clazzname.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def is_foreign_key(fkey):
    '''
    Determines whether a given object is a toco foreign key, or a string containing a classname and the keys necessary to load an object from DynamoDB.

    :param fkey: An object that may or may not be a toco foreign key.
    :rtype: toco object
    '''
    if not isinstance(fkey, str) or not fkey.startswith(FKEY_PREFIX):
        return False
    if fkey in CONSTANT_FKEYS:
        return True
    try:
        obj = json.loads(fkey[len(FKEY_PREFIX):])
        obj['class']
        obj['key']
        return True
    except:
        # If the above throws an exception, we know it isn't a valid foreign key
        # We might want to raise an exception instead if it has the fkey prefix but isn't valid,
        # but I'll add that later if it looks useful.
        return False

def load_from_fkey(fkey, **kwargs):
    '''
    If fkey is a toco foreign key, load it from DynamoDB.

    :param fkey: A toco foreign key.
    :rtype: toco object
    '''
    if not is_foreign_key(fkey):
        return None
    if fkey in CONSTANT_FKEYS:
        return CONSTANT_FKEYS[fkey]
    obj = json.loads(fkey[len(FKEY_PREFIX):])
    key = obj['key']
    clazzname = obj['class']
    del obj['key']
    del obj['class']
    obj.update(key)
    obj.update(**kwargs)
    return get_class(clazzname)._from_fkey(**obj)

def ensure_ddbsafe(d):
    ts = TypeSerializer()
    if isinstance(d, str):
        if len(d) == 0:
            return FKEY_EMPTY_STRING
        else:
            return d
    if isinstance(d, dict):
        return {k:ensure_ddbsafe(d[k]) for k in d}
    elif isinstance(d, list):
        return [ensure_ddbsafe(e) for e in d]
    elif isinstance(d, float):
        return decimal.Decimal(d)
    elif isinstance(d, datetime):
        return d.strftime(DATETIME_FORMAT)
    else:
        return d

def load_constant_fkeys(d):
    if isinstance(d, str) and d in CONSTANT_FKEYS:
        return CONSTANT_FKEYS[d]
    elif isinstance(d, dict):
        return {k:load_constant_fkeys(d[k]) for k in d}
    elif isinstance(d, list):
        return [load_constant_fkeys(e) for e in d]
    else:
        return d

class blob(dict):
    RESERVED_KEYS = ["__predefined_attributes__","__raise_on_miss","_blob__raise_on_miss"]
    def __init__(self, *args, **kwargs):
        __raise_on_miss = bool(kwargs.get("raise_on_miss"))
        if "raise_on_miss" in kwargs:
            del kwargs["raise_on_miss"]
        super().__init__(*args, **kwargs)
        self.__predefined_attributes__ = [a for a in dir(self)]
        self.__predefined_attributes__.append("__predefined_attributes__")
        self.__raise_on_miss = __raise_on_miss
        self.__predefined_attributes__.append("__raise_on_miss")
        for key in self.keys():
            if not key in self.__predefined_attributes__:
                setattr(self, key, self[key])
                pass
            pass
        pass

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if name not in blob.RESERVED_KEYS:
            self.__setitem__(name, value)

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except:
            if object.__getattribute__(self, "__raise_on_miss"):
                return self.__getitem__(name)
            else:
                return self.get(name)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if not key in self.__predefined_attributes__:
            object.__setattr__(self, key, self[key])

    def __delitem__(self, key):
        super().__delitem__(key)
        if not key in self.__predefined_attributes__:
            delattr(self, key)

class BaseTocoObject(object):
    """
    Holder class for a bunch of class methods and stuff like that.
    """
    _SCHEMA_CACHE = None
    _TABLE_CACHE = None
    _CLASSNAME = None
    _REQUIRED_ATTRS = []
    _COMPOUND_ATTRS = {}

    @classmethod
    def _from_dict(cls, d):
        if d.get(JSON_FKEY, None):
            return load_from_fkey(d[JSON_FKEY])
        if d.get(JSON_CLASS, None):
            return get_class(d[JSON_CLASS])(**d)
        return cls(**d)

    @classmethod
    def _encode_nexttoken(cls, key):
        # TODO: replace the json step with something that'll work for any input
        # Convert to a string
        # This won't work for a lot of valid keys due to the json serialization step
        key = json.dumps(key)
        # base64 encode it to make it safer to handle
        key = base64.urlsafe_b64encode(key.encode("utf-8")).decode("utf-8")
        # strip the padding, as we can easily re-add it later
        key = key.replace("=","")
        return key

    @classmethod
    def _decode_nexttoken(cls, key):
        # TODO: replace the json step with something that'll work for any input
        # re-add the padding
        key = key + "=" * ((-1*len(key))%16)
        # convert if back from base64-encoded bytes to the underlying string
        key = base64.urlsafe_b64decode(key.encode("utf-8")).decode("utf-8")
        # load the string back into a dict
        # This won't work for a lot of valid keys due to the json serialization step
        key = json.loads(key)
        return key

    @classmethod
    def _parse_items(cls, response):
        items = []
        for item in response.get("Items",[]):
            params = dict(item)
            params["_in_db"] = True
            params["_attempt_load"] = False
            items.append(cls(**params))
        return items

    @classmethod
    def _preprocess_search_params(cls, **kwargs):
        params = dict(kwargs)
        if params.get("NextToken", None) and not params.get("ExclusiveStartKey", None):
            params["ExclusiveStartKey"] = cls._decode_nexttoken(params["NextToken"])
        if "NextToken" in params:
            del params["NextToken"]
        if not params.get("KeyConditionExpression", None):
            hashname, rangename = cls._HASH_AND_RANGE_KEYS(index_name = params.get("IndexName", None))
            if params.get(hashname) and not params.get("HashKey", None):
                params["HashKey"] = params[hashname]
                del params[hashname]
            if params.get(rangename) and not params.get("RangeKey", None):
                params["RangeKey"] = params[rangename]
                del params[rangename]
            if params.get("HashKey", None) and not params.get("KeyConditionExpression", None):
                hkc = Key(hashname).eq(params["HashKey"])
                if params.get("RangeKey", None):
                    rk = params["RangeKey"]
                    if isinstance(rk,(list, tuple)):
                        rkc = getattr(Key(rangename),rk[0])(*rk[1:])
                    else:
                        rkc = Key(rangename).eq(rk_arg)
                    kce = hkc & rkc
                else:
                    kce = hkc
                params["KeyConditionExpression"] = kce
        if "HashKey" in params:
            del params["HashKey"]
        if "RangeKey" in params:
            del params["RangeKey"]
        return params

    @classmethod
    def _postprocess_search_results(cls, results):
        response = {
            "Items":cls._parse_items(results),
            "NextToken":None,
            "RawResponse":results
        }
        if results.get("LastEvaluatedKey", None):
            response["NextToken"] = cls._encode_nexttoken(results["LastEvaluatedKey"])
        return response

    @classmethod
    def scan(cls, **kwargs):
        params = cls._preprocess_search_params(**kwargs)
        results = cls.TABLE().scan(**params)
        return cls._postprocess_search_results(results)

    @classmethod
    def query(cls, **kwargs):
        params = cls._preprocess_search_params(**kwargs)
        results = cls.TABLE().query(**params)
        return cls._postprocess_search_results(results)

    @classmethod
    def load(cls, **kwargs):
        obj = cls(_attempt_load=True, **kwargs)
        if obj._in_db:
            return obj
        return None

    @classmethod
    def SCHEMA(cls, use_cache=True):
        if cls._SCHEMA_CACHE and use_cache:
            return cls._SCHEMA_CACHE
        schema = cls._SCHEMA()
        cls._SCHEMA_CACHE = schema
        return schema

    @classmethod
    def _SCHEMA(cls, use_cache=True):
        raise NotImplementedError("Each subclass must implement this on their own.")

    @classmethod
    def TABLE_NAME(cls, use_cache=True):
        schema = cls.SCHEMA(use_cache=use_cache)
        return schema.get('TableName')

    @classmethod
    def CLASS_NAME(cls):
        return cls._CLASSNAME if cls._CLASSNAME else "{module}.{name}".format(module=cls.__module__, name=cls.__name__)

    @classmethod
    def TABLE(cls):
        if not cls._TABLE_CACHE:
            cls._TABLE_CACHE = boto3.resource('dynamodb').Table(cls.TABLE_NAME())
        return cls._TABLE_CACHE

    @classmethod
    def create_table(cls):
        boto3.client("dynamodb").create_table(**cls._SCHEMA())

    @classmethod
    def _get_required_attributes(cls):
        attrs = []
        hashkn, rangekn = cls._HASH_AND_RANGE_KEYS()
        if hashkn:
            attrs.append(hashkn)
        if rangekn:
            attrs.append(rangekn)
        attrs.extend(cls._REQUIRED_ATTRS)
        return attrs

    @classmethod
    def _HASH_AND_RANGE_KEYS(cls, index_name=None):
        schema = cls._SCHEMA()
        key_schema = schema['KeySchema']
        if index_name:
            gsis = schema.get("GlobalSecondaryIndexes", [])
            matches = [gsi for gsi in gsis if gsi["IndexName"] == index_name]
            if len(matches) == 0:
                raise RuntimeError("No index with the name '{index_name}' found!".format(index_name=index_name))
            key_schema = matches[0]["KeySchema"]
        hash = [h['AttributeName'] for h in key_schema if h['KeyType']=='HASH'][0]
        ranges = [r['AttributeName'] for r in key_schema if r['KeyType']=='RANGE']
        range = ranges[0] if ranges else None
        return hash, range

    @classmethod
    def _get_class_relation_map(cls, obj):
        return {'class':cls.CLASS_NAME(), 'key':obj._get_key_dict()}

    @classmethod
    def _from_fkey(cls, **kwargs):
        obj = cls(**kwargs)
        obj._needs_reloaded = True
        return obj

    @classmethod
    def _add_compound_attr(cls, attrname, attrfunc, save=False):
        cls._COMPOUND_ATTRS[attrname] = {"func":attrfunc,"save":save}

    @classmethod
    def _remove_compound_attr(cls, attrname):
        if attrname in cls._COMPOUND_ATTRS:
            del cls._COMPOUND_ATTRS[attrname]

class TocoObject(BaseTocoObject):
    '''
    Base class for all DynamoDB-storable toco objects.  Cannot itself be instantiated.

    The only thing that a subclass is required to implement is the classmethod SCHEMA, and it must return a dict that can be passed to client.create_table(**schema) and succeed.

    Constructor args:

    :param kwargs: Keys for an object, and any attributes to attach to that object.
    :rtype: toco object
    '''
    def __init__(self, _in_db=False, _attempt_load=True, **kwargs):
        self._needs_reloaded = False
        self._serialize_as_dict = True
        self._raise_on_getattr_miss = False
        self._obj_dict = blob()
        self._fkey_cache = {}
        self._obj_loaded = {}
        self._obj_updates = {}

        setattr(self, VERSION_KEY, 0)
        self._in_db = _in_db

        if _attempt_load:
            try:
                description = self.__class__.TABLE().get_item(Key=self._get_key_dict(kwargs))
            except ClientError as e:
                description = {}
            if description.get('Item'):
                self._update_attrs(**description['Item'])
                self._clear_update_record()
                self._in_db = True
                # Don't treat init-time changes as real changes if they match the DB.
        self._update_attrs_changed(**kwargs)

    def __setattr__(self, name, value):
        if name.startswith("_"):
            object.__setattr__(self, name, value)
        else:
            if isinstance(value, TocoObject):
                self._obj_dict[name] = value._foreign_key()
                self._fkey_cache[name] = value
            else:
                self._obj_dict[name] = value
                if name in self._fkey_cache:
                    del self._fkey_cache[name]
            if name != VERSION_KEY:
                # The version key is special and shouldn't be tracked
                self._obj_updates[name] = value

    def __getattribute__(self, name):
        if name.startswith("_"):
            # or name in self.__dict__(name not in self._obj_dict and name not in self.__class__._COMPOUND_ATTRS):
            if name in object.__getattribute__(self, "__dict__"):
                return object.__getattribute__(self, "__dict__").get(name)
            return object.__getattribute__(self, name)
        else:
            if self._needs_reloaded:
                self._reload()
                self._needs_reloaded = False
            if name in self._obj_dict:
                value = self._obj_dict[name]
                if is_foreign_key(value):
                    obj = load_from_fkey(value)
                    self._fkey_cache[name] = obj
                    return obj
                else:
                    return load_python_class_if_applicable(self._obj_dict[name])
            elif name in self.__class__._COMPOUND_ATTRS:
                return self.__class__._COMPOUND_ATTRS[name]["func"](self)
            else:
                try:
                    if name in object.__getattribute__(self, "__dict__"):
                        return object.__getattribute__(self, "__dict__").get(name)
                    return object.__getattribute__(self, name)
                except AttributeError as e:
                    if self._raise_on_getattr_miss:
                        raise e
                    else:
                        return None

    def __delattr__(self, name):
        if name in self._obj_dict:
            del self._obj_dict[name]
            if name in self._fkey_cache:
                del self._fkey_cache[name]
            self._obj_updates[name] = None
        else:
            object.__delattr__(self, name)

    def _update_attrs(self, **kwargs):
        kwargs = load_constant_fkeys(kwargs)
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def _update_attrs_changed(self, **kwargs):
        for k in kwargs:
            if getattr(self, k) != kwargs[k]:
                setattr(self, k, kwargs[k])

    def _clear_update_record(self):
        self._obj_loaded = self._obj_dict
        self._obj_updates = {}

    def _get_dict_to_save(self):
        dict_to_save = copy.copy(self._obj_dict)
        compattrs = self.__class__._COMPOUND_ATTRS
        for attrname in compattrs:
            if attrname in dict_to_save:
                continue
            elif compattrs[attrname].get("save", False):
                dict_to_save[attrname] = compattrs[attrname]["func"](self)
        return dict_to_save

    def _get_data_dict(self):
        d = self._get_dict_to_save()
        keys = list(d.keys())
        for k in keys:
            if k.startswith("toco_") or k.endswith("_toco") or "_toco_" in k or k.startswith("_"):
                del d[k]
        return d

    def _my_hash_and_range(self):
        hash_keyname, range_keyname = self.__class__._HASH_AND_RANGE_KEYS()
        hash_key = getattr(self, hash_keyname)
        range_key = getattr(self, range_keyname) if range_keyname else None
        return hash_key, range_key

    def _get_key_dict(self, dictionary=None):
        hash_keyname, range_keyname = self.__class__._HASH_AND_RANGE_KEYS()
        keys = {}
        dictionary = dictionary if dictionary else self._obj_dict
        for k in (hash_keyname, range_keyname):
            if k and k in dictionary.keys():
                # I'm explicitly bypassing the getter here in the off chance either hash or range is a foreign key
                keys[k] = dictionary[k]
        return keys

    def _get_relation_map(self):
        classes = []
        new_classes = [self.__class__]
        has_method = True
        while new_classes:
            bases = []
            for clazz in new_classes:
                if clazz not in classes and hasattr(clazz, "_get_class_relation_map"):
                    classes.append(clazz)
                    bases.extend(clazz.__bases__)
            new_classes = bases
        relation_map = {}
        for clazz in classes[::-1]:
            relation_map.update(clazz._get_class_relation_map(self))
        return relation_map

    def _foreign_key(self):
        '''
        The foreign key necessary to load this object from DynamoDB.
        '''
        return FKEY_PREFIX+json.dumps(self._get_relation_map(), sort_keys=True, separators=(',', ':'))

    @classmethod
    def _json_deserialize(cls, fkey):
        return load_from_fkey(fkey, _attempt_load=False)

    def _json_serialize(self):
        if self._serialize_as_dict:
            d = self._get_dict_to_save()
            d[JSON_CLASS] = self.__class__.CLASS_NAME()
            d[JSON_FKEY] = self._foreign_key()
            return d, "dict"
        else:
            return self._foreign_key(), self.__class__.CLASS_NAME()

    def _save(self, force=False, save_if_missing=True, save_if_existing=True, only_if_updated=False):
        if not save_if_missing and not save_if_existing:
            raise RuntimeError("At least one of save_if_missing and save_if_existing must be true.")

        if only_if_updated and not self._obj_updates:
            return self

        old_version = getattr(self, VERSION_KEY)
        create_condition = Attr(VERSION_KEY).not_exists()
        if force:
            update_condition = Attr(VERSION_KEY).exists()
        else:
            update_condition = Attr(VERSION_KEY).eq(old_version)
        CE = None
        if force and save_if_missing and save_if_existing:
            pass
        elif save_if_missing and save_if_existing:
            CE = Or(create_condition, update_condition)
        elif save_if_existing:
            CE = update_condition
        else:
            # If we're here, we know that create_condition=True
            CE = create_condition
        try:
            setattr(self, VERSION_KEY, old_version+1)
            if CE:
                self._store(CE)
            else:
                self._store()
            self._clear_update_record()
            self._in_db = True
            return self
        except ClientError as e:
            setattr(self, VERSION_KEY, old_version)
            raise e

    def _update(self, force=False):
        return self._save(force=force, save_if_existing=True, save_if_missing=False)

    def _create(self):
        return self._save(force=force, save_if_existing=False, save_if_missing=True)

    def _store(self, CE=None):
        dict_to_save = self._get_dict_to_save()
        required = self._get_required_attributes()
        missing = [r for r in required if not r in dict_to_save or not dict_to_save[r]]
        if missing:
            raise RuntimeError('The following attributes are missing and must be added before saving: '+', '.join(missing))
        dict_to_save = ensure_ddbsafe(dict_to_save)
        if CE:
            self.__class__.TABLE().put_item(Item=dict_to_save, ConditionExpression=CE)
        else:
            self.__class__.TABLE().put_item(Item=dict_to_save)
        return self

    def _delete(self, CE=None):
        if CE:
            return self.__class__.TABLE().delete_item(Key=self._get_key_dict(), ConditionExpression=CE)
        else:
            return self.__class__.TABLE().delete_item(Key=self._get_key_dict())

    def _load(self):
        b = blob()
        b.update(self.__class__.TABLE().get_item(Key=self._get_key_dict()).get("Item", {}))
        # return self.__class__.TABLE().get_item(Key=self._get_key_dict()).get("Item", {})
        return b

    def _reload(self):
        '''
        Reloads the item's attributes from DynamoDB, replacing whatever's currently in the object.
        '''
        self._obj_dict = self._load()
        self._in_db = True
        self._clear_update_record()
        return self

class CFObject(TocoObject):
    '''
    Base class for toco objects that are based on tables created in a CloudFormation stack.
    '''

    _CF_STACK_NAME = None
    _CF_LOGICAL_NAME = None
    _CF_CLIENT = boto3.client('cloudformation')
    _CF_TEMPLATE = None
    _CF_RESOURCES = {}

    @classmethod
    def _set_cf_info(cls, cf_stack_name=None, cf_logical_name=None):
        if cf_stack_name:
            cls._CF_STACK_NAME = cf_stack_name
        if cf_logical_name:
            cls._LOGICAL_NAME = cf_logical_name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _get_stack_name(cls, stack_name=None):
        stack_name = stack_name if stack_name else cls._CF_STACK_NAME
        if not stack_name:
            raise RuntimeError("Stack name not set!")
        return stack_name

    @classmethod
    def _get_stack_and_logical_names(cls, stack_name=None, logical_name=None):
        stack_name = stack_name if stack_name else cls._CF_STACK_NAME
        logical_name = logical_name if logical_name else cls._CF_LOGICAL_NAME
        if not stack_name or not logical_name:
            raise RuntimeError("Stack name or logical name (or both) not set!")
        return stack_name, logical_name

    @classmethod
    def _describe_stack_resource(cls, stack_name=None, logical_name=None):
        stack_name, logical_name = cls._get_stack_and_logical_names(stack_name=stack_name, logical_name=logical_name)
        if logical_name not in cls._CF_RESOURCES:
            logging.warn("Cache miss loading _CF_RESOURCE for class {}".format(cls))
            response = cls._CF_CLIENT.describe_stack_resource(StackName=stack_name, LogicalResourceId=logical_name)
            if not response or "StackResourceDetail" not in response:
                raise RuntimeError("Resource does not exist!")
            cls._CF_RESOURCES[logical_name] = response["StackResourceDetail"]
        else:
            logging.info("Cache hit loading _CF_RESOURCE for class {}".format(cls))
        return cls._CF_RESOURCES[logical_name]

    @classmethod
    def _get_physical_resource_id(cls, stack_name=None, logical_name=None):
        return cls._describe_stack_resource(stack_name=stack_name, logical_name=logical_name)["PhysicalResourceId"]

    @classmethod
    def _get_template(cls, stack_name=None):
        stack_name = cls._get_stack_name(stack_name=stack_name)
        if not getattr(cls, "_CF_TEMPLATE"):
            logging.warn("Cache miss loading _CF_TEMPLATE for class {}".format(cls))
            try:
                template = cls._CF_CLIENT.get_template(StackName=stack_name)["TemplateBody"]
                setattr(cls, "_CF_TEMPLATE", template)
            except ValidationError:
                raise RuntimeError("Unable to retrieve template for stack {}, likely due to it not existing.".format(stack_name))
        else:
            logging.info("Cache hit loading _CF_TEMPLATE for class {}".format(cls))
        return cls._CF_TEMPLATE

    @classmethod
    def _clear_cf_cache(cls):
        setattr(cls, "_CF_TEMPLATE", None)
        setattr(cls, "_CF_RESOURCES", {})

    @classmethod
    def _SCHEMA(cls):
        stack_name, logical_name = cls._get_stack_and_logical_names()
        template = cls._get_template(stack_name=stack_name)
        resources = template["Resources"]
        if logical_name not in resources:
            raise RuntimeError("Stack doesn't contain a table with the given logical name!")
        resource = resources[logical_name]
        table_type = "AWS::DynamoDB::Table"
        if not table_type == resource["Type"]:
            raise RuntimeError("Logical resource {} in stack {} is of type '{}', not type '{}'".format(logical_name, stack_name, resource["Type"], table_type))
        properties = dict(resource["Properties"])
        properties["TableName"] = cls._get_physical_resource_id()
        return properties

    @classmethod
    def _get_class_relation_map(cls, obj):
        stack_name, logical_name = obj._get_stack_and_logical_names()
        return {'class':cls.CLASS_NAME(), '_cf_stack_name':stack_name, '_cf_logical_name':logical_name}

    @classmethod
    def _from_fkey(cls, _cf_stack_name, _cf_logical_name, **kwargs):
        return cls.lazysubclass(stack_name=_cf_stack_name, logical_name=_cf_logical_name)(**kwargs)

    @classmethod
    def lazysubclass(cls, stack_name=None, logical_name=None):
        '''
        Returns a class that inherits from this one, with the given default stack and logical names.
        If you just want to have a new object type with no fancy features added, this makes it a one-liner.

        :rtype: Class that inherits from cls
        '''
        class LazyObject(cls):
            _CF_STACK_NAME = stack_name if stack_name else cls._CF_STACK_NAME
            _CF_LOGICAL_NAME = logical_name if logical_name else cls._CF_LOGICAL_NAME
            _CLASSNAME = cls.CLASS_NAME()
        return LazyObject



#!/usr/bin/env python

# https://misc.flogisoft.com/bash/tip_colors_and_formatting
SHELL_COLORS = {
    'RED':'\033[91m',
    'GREEN':'\033[92m',
    'YELLOW':'\033[93m',
    'BLUE':'\033[94m',
    'PURPLE':'\033[95m',
    'WHITE':'\033[97m',
    'RESET':'\033[0m',
}

def color_text(s, c):
    return '{}{}{}'.format(SHELL_COLORS.get(c.upper(), SHELL_COLORS["BLUE"]), s, SHELL_COLORS["RESET"])

def blue_text(s):
    return color_text(s, "BLUE")

def print_blue(s):
    return print(blue_text(s))

def white_text(s):
    return color_text(s, "WHITE")

def print_white(s):
    return print(white_text(s))

def purple_text(s):
    return color_text(s, "PURPLE")

def print_purple(s):
    return print(purple_text(s))

def green_text(s):
    return color_text(s, "GREEN")

def print_green(s):
    return print(green_text(s))

def yellow_text(s):
    return color_text(s, "YELLOW")

def print_yellow(s):
    return print(yellow_text(s))

def red_text(s):
    return color_text(s, "RED")

def print_red(s):
    return print(red_text(s))
